
package trafficlighttest;

public class Intersection {
    
    //Names of the lanes
    private static final int NORTHLEFT = 0;
    private static final int SOUTHLEFT = 1;
    private static final int NORTH = 2;
    private static final int SOUTH = 3;
    private static final int EASTLEFT = 4;
    private static final int WESTLEFT = 5;
    private static final int EAST = 6;
    private static final int WEST = 7;
   
    //Names of the modes, NSL = north and south left turning, NS = non-left turning north and south traffic
    //EWL = east and west left turning, EW = east and west non-left turning traffic
    private static final int NSL = 0;
    private static final int NS = 1;
    private static final int EWL = 2;
    private static final int EW = 3;
   
    //Array to hold the number of cars in each lane
    protected int[] carsInEachLane;
   
    //Array to hold the lanes of the intersection
    protected Lane[] lanes;
   
    //Array to hold the rules that govern light changes
    protected Rule[] rules;
   
    //Array to hold the different modes the light/intersection can be in
    protected State[] modes;
   
    //Array to hold the average wait time of cars in each lane
    protected int[] averageWaits;
   
    //Array to hold the worst wait time of each lane
    protected int[] worstWaits;
    
    //Array to hold the total cars from each lane
    protected int[] totalCars;
   
    //The current mode/state we are in
    protected int currentMode;
   
    //The most recently used rule, this is used to decide which rule is looked at next
    //Works round robin style, each rule will get a chance to change the light before any 
    //rule gets a second turn
    protected int currentRule;
   
    //The number of cycles since the last time the light changed
    protected int timeSinceLastChange;
   
    //The number of total rules, helps with the round robin style rule checking
    protected int numRules;

    /* Constructor
     * Creates an intersection with the specified lanes
     * allLanes = the array of lanes, a 1 = yes create that lane, any other number = no don't create that lane
     * allows for customizable number and type of lanes in the intersection, each spot in the array is a unique lane
     */
    public Intersection(int[] allLanes) {
        this.lanes = new Lane[8];
        for (int i = 0; i < allLanes.length; i++) {
            if (allLanes[i] == 1) {
                this.lanes[i] = new Lane(i);
            }
            else {
                this.lanes[i] = null;
            }
        }
        this.rules = new Rule[1];
        this.carsInEachLane = new int[8];
        this.averageWaits = new int[8];
        this.worstWaits = new int[8];
        this.totalCars = new int[8];
        
        this.modes = new State[4];
        this.modes[0] = new NSL();
        this.modes[1] = new NS();
        this.modes[2] = new EWL();
        this.modes[3] = new EW();
        
        this.currentMode = 0;
        this.currentRule = 0;
        this.timeSinceLastChange = 0;
        this.numRules = 0;
    }
    
    /* updateCarCounts
     * Grabs the current number of cars in each lane and puts them in the intersection's array
     */
    private void updateCarCounts() {
        for (int i = 0; i < this.carsInEachLane.length; i++) {
            this.carsInEachLane[i] = this.lanes[i].getCurrentCars();
        }
    }
    
    /* getCarCounts
     * Updates the car count of each lane and returns the car count of each lane as an array
     */
    public int[] getCarCounts() {
        updateCarCounts();
        return carsInEachLane;
    }
    
    /* getAverageWaits
     * Retrieves the average wait time of each lane from each lane and returns the results as an array
     */
    public int[] getAverageWaits() {
        for (int i = 0; i < this.averageWaits.length; i++) {
            this.averageWaits[i] = lanes[i].getAverageWait();
        }
        return averageWaits;
    }
    
    /* getWorstWaits
     * Retrieves the worst wait time for a car in each lane and returns the results as an array
     */
    public int[] getWorstWaits() {
        for (int i = 0; i < worstWaits.length; i++) {
            this.worstWaits[i] = lanes[i].getWorstWait();
        }
        return worstWaits;
    }
   
    /* addRule
     * Adds a rule to the set of rules by extending the rule array one slot
     * lane = the lane that the rule governs
     * allowed = the maximum cars allowed in said lane before they get a green
     */
    public void addRule(int lane, int allowed){
        if (numRules == 0){
            rules[0] = new Rule(lane, allowed);
            numRules++;
            return;
        }
        numRules++;
        Rule[] temp = new Rule[numRules];
        for (int i = 0; i < rules.length; i++) {
            temp[i] = rules[i];
        }
        temp[temp.length - 1] = new Rule(lane, allowed);
        this.rules = new Rule[numRules];
        for (int i = 0; i < temp.length; i++) {
            rules[i] = temp[i];
        }       
    }
    
    /* step
     * This is the master controller of the simulation. Step actions in order:
     * 1) Add a new car to a random lane
     * 2) Check the rules to see if a new mode should be entered
     * 3) If no to number 2, check to see if we should enter the next mode based on time since last change
     * 4) If no to number 3, increment the appropriate wait times and time since last change
     */
    public void step() {
        //1
        System.out.println("Adding a car to a random lane...");
        int random = (int)(Math.random() * 8);
        lanes[random].addCar(currentMode);
        int[] now = this.getCarCounts();
        System.out.println("Current car counts: ");
        for (int i = 0; i < now.length; i++) {
            System.out.printf("Lane %d:%d\n", i, now[i]);
        }
        int test = checkRules();
        int test2 = checkTimer();
        //2
        if (currentMode != test) {
            currentMode = test;
            sendCars();
            System.out.println();
        }
        else if(currentMode != test2) { //3
            currentMode = test2;
            sendCars();
            System.out.println();
        }
        else { //4
            incrementWaits();
            timeSinceLastChange++;
            System.out.println();
        }
    }
    
    /* checkRules
     * Checks the rules to see if there are enough cars waiting in a lane to warrant a green.
     * If so, returns the new mode that should be entered, otherwise returns the mode we started in.
     */
    private int checkRules() {
        int newMode = currentMode;
        for (int i = currentRule; i < numRules + currentRule; i++) {
            if (lanes[rules[i % numRules].getLane()] != null) {
                if (lanes[rules[i % numRules].getLane()].getCurrentCars() > rules[i % numRules].getAllowed()){
                   int lane = rules[i % numRules].getLane();
                   newMode = lane / 2;
                   System.out.printf("Rule based change: Entering mode %d...\n", newMode);
                   timeSinceLastChange = 0;
                   currentRule = i % numRules;
                   break;
                }//end rule check if
            }//end lane existence check if    
        }//end for
        return newMode;
    }
    
    /* checkTimer
     * Checks to see if the time since last change is at its limit.
     * If yes, return the new mode we should enter, otherwise return the mode we started in.
     */
    private int checkTimer() {
        int newMode = currentMode;
        if (timeSinceLastChange == 10) {
            timeSinceLastChange = 0;
            newMode = (currentMode + 1) % 4;
            System.out.printf("10 cycles since a change: Entering mode %d...\n", newMode);
        }
        return newMode;
    }
    
    /* sendCars
     * Calls the appropriate mode's/state's send car function which sends cars through the intersection 
     */
    private void sendCars() {
        modes[currentMode].sendCars(lanes);
    }
    
    /* incrementWaits
     * Calls the appropriate mode's/state's increment waits function which tells all lanes that don't have a green
     * to increment their cars' wait times
     */
    private void incrementWaits() {
        modes[currentMode].incrementWaits(lanes);
    }
    
    public int[] getTotalCars() {
        for (int i = 0; i < totalCars.length; i++) {
            totalCars[i] = lanes[i].getTotalCars();
        }
        return totalCars;
    }
}
